/*
 * copyright (c) 1999 Rob Ross and Phil Carns, all rights reserved.
 *
 * Written by Rob Ross and Phil Carns, funded by Scyld Computing.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Contact:  Rob Ross    rbross@parl.clemson.edu
 *           Phil Carns  pcarns@parl.clemson.edu
 */

#define __NO_VERSION__

#include "config.h"
#include "pvfs_kernel_config.h"
#include "pvfs_linux.h"
#include "ll_pvfs.h"
#include "pvfs_proc.h"

#ifdef HAVE_LINUX_LOCKS_H
#include <linux/locks.h>
#endif
#ifdef HAVE_LINUX_SMP_LOCK_H
#include <linux/smp_lock.h>
#endif

/* dentry ops */
static int pvfs_dentry_revalidate(struct dentry *de, int);

/* dir inode-ops */
static struct dentry *pvfs_lookup(struct inode *dir, struct dentry *target);
static int pvfs_unlink(struct inode *dir_inode, struct dentry *entry);
static int pvfs_mkdir(struct inode *dir_inode, struct dentry *entry, int mode);
static int pvfs_rmdir(struct inode *dir_inode, struct dentry *entry);

/* dir file-ops */
static int pvfs_readdir(struct file *file, void *dirent, filldir_t filldir);
static ssize_t pvfs_dir_read(struct file *file, char *buf, size_t count,
	loff_t *ppos);

struct dentry_operations pvfs_dentry_operations =
{
	d_revalidate:     pvfs_dentry_revalidate
};

struct inode_operations pvfs_dir_inode_operations =
{
	/* put this in right later... pvfs_notify_change */
	/* &pvfs_dir_operations, */
	create: pvfs_file_create,    /* create TODO: DOES THIS NEED TO BE HERE? */
	lookup: pvfs_lookup,      /* lookup */
	unlink: pvfs_unlink, /* unlink called on parent directory to unlink a file */
	mkdir: pvfs_mkdir,           /* mkdir */
	rmdir: pvfs_rmdir,              /* rmdir */
	rename: pvfs_rename,         /* rename */
	revalidate: pvfs_revalidate_inode,   /* revalidate */
	setattr: pvfs_notify_change
};

struct file_operations pvfs_dir_operations = {
	read: pvfs_dir_read,   /* read -- returns -EISDIR  */
	readdir: pvfs_readdir  /* readdir */
};

/* inode operations for directories */

/* pvfs_lookup(dir, entry)
 *
 * Notes:
 * - name to look for is in the entry (entry->d_name.name, entry->d_name.len)
 * - dir is the directory in which we are doing the lookup
 *
 * Behavior:
 * - call ll_pvfs_lookup to get a handle
 * - grab an inode for the new handle
 * - fill in the inode metadata
 * - add the dentry/inode pair into the dcache
 * - (optionally) set up pointer to new dentry functions
 *
 * Returns NULL pretty much all the time.  I know, this seems really
 * screwed up, but the ext2 fs seems to be doing the same thing...
 */
static struct dentry *pvfs_lookup(struct inode *dir, struct dentry *entry)
{
	int error = 0, len_dir, len_file;
	struct pvfs_inode *pinode, *parent_pinode = NULL;
	struct pvfs_meta meta;
	struct inode *inode = NULL;
	char *ptr;
	/* update the statistics */
	if(pvfs_collect_stats) pvfs_vfs_stat.lookup++;
	PENTRY;
	parent_pinode = pvfs_inop(dir);

	if(read_lock_inode(parent_pinode) < 0) {
			  return NULL;
	}

	PDEBUG(D_DIR, "pvfs_lookup called on %ld for %s\n", dir->i_ino,
	entry->d_name.name);
	
	PDEBUG(D_DIR, "name might be %s/%s\n", pvfs_inop(dir)->name,
	entry->d_name.name);

	len_dir = strlen(pvfs_inop(dir)->name);
	len_file = entry->d_name.len;
	if ((pinode = (struct pvfs_inode *) kmalloc(sizeof(struct pvfs_inode)
	+ len_dir + len_file + 2, GFP_KERNEL)) == NULL){
		read_unlock_inode(parent_pinode);
		PEXIT;
		return NULL;
	}

	/* fill in pvfs_inode name field first */
	init_pvfs_rwsem(&pinode->lock);
	pinode->name = (int8_t *)pinode + sizeof(struct pvfs_inode);
	ptr = pinode->name;
	strcpy(ptr, pvfs_inop(dir)->name);
	ptr += len_dir;
	*ptr = '/';
	ptr++;
	strcpy(ptr, entry->d_name.name);

	/* do the lookup, grabs metadata */
	error = ll_pvfs_lookup(pvfs_sbp(dir->i_sb), pinode->name,
		len_dir + len_file + 1, &meta);
	if (error < 0 && error != -ENOENT) {
		/* real error */
		kfree(pinode);
		read_unlock_inode(parent_pinode);
		PEXIT;
		return ERR_PTR(error);
	}

	if (error == -ENOENT) {
		/* no file found; need to insert NULL inode so that create calls
		 * can work correctly (or so says Gooch)
		 */
		entry->d_time = 0;
		entry->d_op = &pvfs_dentry_operations;
		d_add(entry, NULL);
		kfree(pinode);
		read_unlock_inode(parent_pinode);
		PEXIT;
		return NULL;
	}

	if ((inode = iget(dir->i_sb, meta.handle)) == NULL) {
		/* let's drop the dentry here... */
		kfree(pinode);
		read_unlock_inode(parent_pinode);
		PEXIT;
		return ERR_PTR(-ENOMEM);
	}

	/* fill in inode structure, remainder of pvfs_inode */
	pinode->handle = inode->i_ino;
	pinode->super = pvfs_sbp(inode->i_sb);
	pvfs_meta_to_inode(&meta, inode);
	inode->u.generic_ip = pinode;

	entry->d_time = 0;
	entry->d_op = &pvfs_dentry_operations;

	PDEBUG(D_DIR, "saved name is %s\n", pvfs_inop(inode)->name);

	if (S_ISDIR(inode->i_mode)) {
		inode->i_op = &pvfs_dir_inode_operations;
		inode->i_fop = &pvfs_dir_operations;
	}
   else if (S_ISREG(inode->i_mode)) {
		inode->i_op = &pvfs_file_inode_operations;
		inode->i_fop = &pvfs_file_operations;
		inode->i_data.a_ops = &pvfs_file_aops;
	}
   else {
		inode->i_op = NULL;
		inode->i_fop = NULL;
	}

	d_add(entry, inode);
	read_unlock_inode(parent_pinode);

	PEXIT;
	return NULL;
}


/* pvfs_mkdir()
 *
 * Called on parent directory with dentry of directory to create.
 *
 * This is pretty much a carbon copy of pvfs_file_create().
 */
static int pvfs_mkdir(struct inode *dir, struct dentry *entry, int mode)
{
	int error = 0, len_dir, len_file;
	struct inode *inode;
	struct pvfs_meta meta;
	struct pvfs_inode *pinode, *parent_pinode = NULL;
	char *ptr;
	/* update the statistics */
	if(pvfs_collect_stats) pvfs_vfs_stat.mkdir++;    
	PENTRY;
	parent_pinode = pvfs_inop(dir);
	if((error = write_lock_inode(parent_pinode)) < 0) {
		 return error;
	}

	PDEBUG(D_DIR, "pvfs_mkdir called on %ld for %s\n", dir->i_ino,
	entry->d_name.name);

	len_dir = strlen(pvfs_inop(dir)->name);
	len_file = entry->d_name.len;
	if ((pinode = (struct pvfs_inode *) kmalloc(sizeof(struct pvfs_inode)
	+ len_dir + len_file + 2, GFP_KERNEL)) == NULL){
		write_unlock_inode(parent_pinode);
		PEXIT;
		return -ENOMEM;
	}

	/* build pvfs_inode name field first */
	init_pvfs_rwsem(&pinode->lock);
	/* build pvfs_inode name field first */
	pinode->name = (int8_t *)pinode + sizeof(struct pvfs_inode);
	ptr = pinode->name;
	strcpy(ptr, pvfs_inop(dir)->name);
	ptr += len_dir;
	*ptr = '/';
	ptr++;
	strcpy(ptr, entry->d_name.name);

	/* do the create */
	meta.valid = V_MODE | V_UID | V_GID | V_TIMES;
	meta.uid = current->fsuid;
	meta.gid = current->fsgid;
	meta.mode = mode;
	meta.mtime = meta.atime = meta.ctime = CURRENT_TIME;

	PDEBUG(D_DIR, "pvfs_mkdir calling ll_pvfs_mkdir\n");
	if ((error = ll_pvfs_mkdir(pvfs_sbp(dir->i_sb), pinode->name,
	len_dir + len_file + 1, &meta )) < 0)
	{
		kfree(pinode);
		write_unlock_inode(parent_pinode);
		PEXIT;
		return error;
	}

	/* do a lookup so we can fill in the inode */
	PDEBUG(D_DIR, "pvfs_mkdir calling ll_pvfs_lookup\n");
	if ((error = ll_pvfs_lookup(pvfs_sbp(dir->i_sb), pinode->name,
	len_dir + len_file + 1, &meta)) < 0)
	{
		kfree(pinode);
		write_unlock_inode(parent_pinode);
		PEXIT;
		return error;
	}

	/* fill in inode structure and remainder of pvfs_inode */
	if ((inode = iget(dir->i_sb, meta.handle)) == NULL) {
		kfree(pinode);
		write_unlock_inode(parent_pinode);
		PEXIT;
		return -ENOMEM;
	}

	pinode->handle = inode->i_ino;
	pinode->super = pvfs_sbp(inode->i_sb);
	pvfs_meta_to_inode(&meta, inode);
	inode->u.generic_ip = pinode;

	d_instantiate(entry, inode); /* do I know what this does?  nope! */

	PDEBUG(D_DIR, "saved name is %s\n", pvfs_inop(inode)->name);

	if (S_ISDIR(inode->i_mode)) {
		inode->i_op = &pvfs_dir_inode_operations;
		inode->i_fop = &pvfs_dir_operations;
	}
   else if (S_ISREG(inode->i_mode)) {
		inode->i_op = &pvfs_file_inode_operations;
		inode->i_fop = &pvfs_file_operations;
		inode->i_data.a_ops = &pvfs_file_aops;
	}
   else {
		inode->i_op = NULL;
		inode->i_fop = NULL;
	}

	write_unlock_inode(parent_pinode);
	PEXIT;
	return error;
}

/* pvfs_unlink(dir, entry)
 *
 * Called on parent directory with dentry of file to unlink.  We call
 * d_delete(entry) when done to take the dentry out of the cache.  This
 * call also frees the inode associated with the unlinked file.
 * 
 */
int pvfs_unlink(struct inode *dir, struct dentry *entry)
{
	int error = 0;
	struct pvfs_inode *parent_pinode = NULL;
	/* update the statistics */
	if(pvfs_collect_stats) pvfs_vfs_stat.unlink++;   
	PENTRY;
	parent_pinode = pvfs_inop(dir);
	if((error = write_lock_inode(parent_pinode)) < 0) {
		 return error;
	}

	PDEBUG(D_DIR, "pvfs_unlink called on %ld for %s\n", dir->i_ino,
	entry->d_name.name);
	if (entry->d_inode == NULL) {
		/* negative dentry ?!? */
		PERROR("pvfs_unlink: negative dentry?\n");
		/* TODO: should look up the file? */
		write_unlock_inode(parent_pinode);
		PEXIT;
		return -ENOENT;
	}
	if (entry->d_inode->i_nlink == 0) {
		PERROR("pvfs_unlink: deleting nonexistent file?\n");
	}

	error = ll_pvfs_unlink(pvfs_inop(entry->d_inode));

	if (error == 0) {
		entry->d_inode->i_nlink--;	
	}
	write_unlock_inode(parent_pinode);

	PEXIT;
	return error;
}


/* pvfs_rmdir()
 *
 * Called on parent directory with dentry of directory to remove.
 *
 * TODO: I'm not so sure that I know everything this call should be
 * doing; look at minix/namei.c:minix_rmdir() for some possible ideas.
 */
int pvfs_rmdir(struct inode *dir, struct dentry *entry)
{
	int error = 0;
	struct pvfs_inode *parent_pinode = NULL;
	/* update the statistics */
	if(pvfs_collect_stats) pvfs_vfs_stat.rmdir++;       
	PENTRY;
	parent_pinode = pvfs_inop(dir);
	if((error = write_lock_inode(parent_pinode)) < 0) {
		 return error;
	}

	if (!d_unhashed(entry))
	{
		PDEBUG(D_DIR, "pvfs_rmdir failing because dir (%s) is busy.\n", 
			entry->d_name.name);
		write_unlock_inode(parent_pinode);
		return -EBUSY;
	}

	PDEBUG(D_DIR, "pvfs_rmdir called on %ld for %s\n", dir->i_ino,
	entry->d_name.name);
	if (entry->d_inode == NULL) {
		/* negative dentry ?!? */
		PERROR("pvfs_rmdir: negative dentry?\n");
		/* TODO: should look up the file? */
		write_unlock_inode(parent_pinode);
		PEXIT;
		return -ENOENT;
	}
	if (entry->d_inode->i_nlink == 0) {
		PERROR("pvfs_rmdir: deleting nonexistent directory?\n");
	}

	error = ll_pvfs_rmdir(pvfs_inop(entry->d_inode));

	if (error == 0)
	{
		entry->d_inode->i_nlink--;	
	}
	write_unlock_inode(parent_pinode);

	PEXIT;
	return error;
}


/* pvfs_rename(old_dir, old_dentry, new_dir, new_dentry)
 *
 * PARAMETERS:
 * old_dir - inode of parent directory of file/dir to be moved
 * new_dir - inode of parent directory of destination
 * old_dentry - dentry referring to the file/dir to move
 * new_dentry - dentry referring to the destination; inode might or
 *   might not be available.
 *
 * NOTES:
 * If the target name already exists, there will be a valid inode in the
 * new_dentry.  If the inode pointer is NULL, then there is nothing
 * referred to by that name.
 *
 * Also need to update the dcache and such to take changes into account!
 *
 * see fs/nfs/dir.c:nfs_rename() for an example.
 * see fs/minix/namei.c:minix_rename() for a different example.
 *
 * It's not enough to just mark the various inodes as dirty.  We need to
 * ensure that the directory cache is updated.  This turns out to be
 * easy; just call d_move(old_dentry, new_dentry) to get the dentries
 * updated!  Or at least, that seems to be working <smile>...
 *
 * Returns 0 on success, -errno on failure.
 */
int pvfs_rename(struct inode *old_dir, struct dentry *old_dentry, 
struct inode *new_dir, struct dentry *new_dentry)
{
	int error = -ENOSYS;
	struct inode *new_inode;
	struct pvfs_inode *old_pinode, *new_pinode;
	struct pvfs_inode *old_parent_pinode = NULL, *new_parent_pinode = NULL;
	/* update the statistics */
	if(pvfs_collect_stats) pvfs_vfs_stat.rename++;   
	PENTRY;

	old_parent_pinode = pvfs_inop(old_dir);
	new_parent_pinode = pvfs_inop(new_dir);
	if((error = write_lock_inode(old_parent_pinode)) < 0) {
		 return error;
	}
	if(new_parent_pinode != old_parent_pinode) {
		if((error = write_lock_inode(new_parent_pinode)) < 0) {
			 write_unlock_inode(old_parent_pinode);
			 return error;
		}
	}
	old_pinode = pvfs_inop(old_dentry->d_inode);
	new_inode = new_dentry->d_inode;
	if (new_inode == NULL) {
		/* no file at new name; need to create a fake pvfs_inode */
		/* namebuf is big enough for <host>:port/<name> */
		char *namebuf;
		struct pvfs_inode fake_pinode;
		init_pvfs_rwsem(&fake_pinode.lock);

		namebuf = (char *) kmalloc(PVFSHOSTLEN + PVFSNAMELEN + 8, GFP_KERNEL);
		if (namebuf == NULL) {
			  if(new_parent_pinode != old_parent_pinode) {
				  write_unlock_inode(new_parent_pinode);
			  }
			  write_unlock_inode(old_parent_pinode);
			  return 0;
		}

		PDEBUG((D_DIR | D_FILE), "pvfs_rename, no inode for new file\n");
		if ((strlen(old_pinode->name) + new_dentry->d_name.len + 2) >
		PVFSHOSTLEN + PVFSDIRLEN + 7) {
			/* should never happen */
			PERROR("pvfs_rename: string too long?!?\n");
			PEXIT;
			kfree(namebuf);
			if(new_parent_pinode != old_parent_pinode) {
				write_unlock_inode(new_parent_pinode);
			}
	 		write_unlock_inode(old_parent_pinode);
			return -ENOSYS;
		}
		sprintf(namebuf, "%s/%s", pvfs_inop(new_dir)->name,
		new_dentry->d_name.name);
		fake_pinode.handle = 0;
		fake_pinode.name = namebuf;
		fake_pinode.super = pvfs_sbp(new_dir->i_sb);

		PDEBUG((D_DIR | D_FILE), "pvfs_rename called, %s -> %s (new)\n",
		old_pinode->name, namebuf);

		error = ll_pvfs_rename(old_pinode, &fake_pinode);
		kfree(namebuf);
	}
	else {
		/* a file already exists with new name */
		new_pinode = pvfs_inop(new_dentry->d_inode);
		PDEBUG((D_DIR | D_FILE), "pvfs_rename called, %s -> %s\n",
		old_pinode->name, new_pinode->name);

		error = ll_pvfs_rename(old_pinode, new_pinode);
	}
	
	/* Update the dcache */
	d_move(old_dentry, new_dentry);
	if(new_parent_pinode != old_parent_pinode) {
		write_unlock_inode(new_parent_pinode);
	}
 	write_unlock_inode(old_parent_pinode);

	PEXIT;
	return error;
}

/* pvfs_readdir()
 *
 */
int pvfs_readdir(struct file *file, void *dirent, filldir_t filldir)
{
	int ret, len, i;
	struct pvfs_dirent *dbuf = NULL;
	struct inode *inode;
	pvfs_off_t pos;
	struct pvfs_inode *parent_pinode = NULL;
	/* update the statistics */
	if(pvfs_collect_stats) pvfs_vfs_stat.readdir++;   
	PENTRY;

	/* FIXME: Right now we are using FETCH_DENTRY_COUNT * ~1K of memory for this buffer */

	dbuf = (struct pvfs_dirent *)vmalloc(FETCH_DENTRY_COUNT * sizeof(struct pvfs_dirent));
	if(!dbuf) {
		PERROR("Could not allocate dbuf for process %d\n",(int)current->pid);
		return -ENOMEM;
	}

	inode = file->f_dentry->d_inode;
	parent_pinode = pvfs_inop(inode);
	if((ret = read_lock_inode(parent_pinode)) < 0) {
		 vfree(dbuf);
		 return ret;
	}

	PDEBUG(D_DIR, "pvfs_readdir called for %ld.\n", inode->i_ino);

	pos = file->f_pos;
	while(1) {
		int count;
		if ((ret = ll_pvfs_readdir(pvfs_inop(inode), dbuf, &pos, FETCH_DENTRY_COUNT)) < 0)
			goto out;  /* error */
		count = ret;

		for(i=0; i < count; i++) {
			len = strlen(dbuf[i].name);
			/* glibc did not give us sufficient buffer space */
			if ((ret = filldir(dirent, dbuf[i].name, len, dbuf[i].off, dbuf[i].handle, DT_UNKNOWN))) {
				file->f_pos = dbuf[i].off;
				ret = 0;
				goto out;
			}
		}
		/* ll_pvfs_readdir gives us the position to start from the next time */
		file->f_pos = pos;
		
		/* eof */
		if(count != FETCH_DENTRY_COUNT) {
			ret = 0;
			goto out;
		}
	}
out:
	read_unlock_inode(parent_pinode);
	vfree(dbuf);
	PEXIT;
	return ret;
}

/* pvfs_fsync(file, dentry, [datasync])
 *
 */
int pvfs_fsync(struct file *file, struct dentry *dentry, int datasync)
{
	int error = 0;
	struct pvfs_inode *pinode;
	/* update the statistics */
	if(pvfs_collect_stats) pvfs_vfs_stat.fsync++; 
	PENTRY;

	lock_kernel();
	pinode = pvfs_inop(dentry->d_inode);

	error = ll_pvfs_fsync(pinode);
	unlock_kernel();

	PEXIT;
	return error;
}

/* pvfs_dir_read()
 *
 * Returns -EISDIR.  Just here to get the right error value back.
 */
static ssize_t pvfs_dir_read(struct file *file, char *buf, size_t count,
loff_t *ppos)
{
	PENTRY;
	PEXIT;
	return -EISDIR;
}

/* pvfs_revalidate_inode()
 *
 * Performs a getmeta operation to ensure that the data in the inode is
 * up-to-date.
 *
 * QUESTION: Do we need a lock on the inode in here?
 */
int pvfs_revalidate_inode(struct dentry *dentry)
{
	int error = 0, len_dir, len_file;
	struct inode *inode;
	struct dentry *parent_dentry;
	struct inode *parent_inode;
	struct pvfs_inode *pinode, *parent_pinode = NULL;
	struct pvfs_meta meta;
	struct pvfs_phys phys;
	pvfs_handle_t old_handle;
	umode_t old_mode;
	char *ptr;
	/* update the statistics */
	if(pvfs_collect_stats) pvfs_vfs_stat.revalidate++;    
	PENTRY;

	lock_kernel();

	inode = dentry->d_inode;
	if (inode == NULL) {
		PERROR("pvfs_revalidate_inode called for NULL inode\n");
		unlock_kernel();
		PEXIT;
		return -EINVAL;
	}

#ifdef IS_DEADDIR
	if(IS_DEADDIR(dentry->d_inode))
	{
		PERROR("DEAD directory detected going into pvfs_revalidate_inode.\n");
	}
#endif

	/* get the parent dentry and inode -- this should be optimized later */
	parent_dentry = dentry->d_parent;
	if (parent_dentry == NULL) {
		PERROR("pvfs_revalidate_inode: d_parent is NULL\n");
		unlock_kernel();
		PEXIT;
		return -EINVAL;
	}

	parent_inode = parent_dentry->d_inode;
	if (parent_inode == NULL) {
		PERROR("pvfs_revalidate_inode: parent inode is NULL\n");
		unlock_kernel();
		PEXIT;
		return -EINVAL;
	}
	parent_pinode = pvfs_inop(parent_inode);
	if((error = read_lock_inode(parent_pinode)) < 0) {
		unlock_kernel();
		PEXIT;
		return error;
	}
	
	/* reconstruct the name in the inode */
	len_dir = strlen(pvfs_inop(parent_inode)->name);
	len_file = dentry->d_name.len;
	pinode = pvfs_inop(inode);

	/* save the old handle value for debugging test */
	old_handle = pinode->handle;
	old_mode = inode->i_mode;

	/* if we're not looking at the root of a file system, we need to 
	 * rebuild the filename.  it's important that pinode point to
	 * pvfs_inop(inode) at this point in case we don't go in the "if".
	 */
	if (parent_inode != inode) {
		/* the "1" takes into account the '/' between file and dir names
		 *
		 * if the old name is longer or of equal length, we don't realloc.
		 */
		if (strlen(pinode->name) < len_file + len_dir + 1) {
			struct pvfs_inode *old_pinode;

			/* we need to reallocate the pvfs_inode structure and file name.
			 *
			 * the "2" takes into account the '/' between names and the
			 * trailing '\0'.
			 */
         old_pinode = pinode;
			pinode = (struct pvfs_inode *) kmalloc(sizeof(struct pvfs_inode)
																+ len_dir + len_file + 2, 
																GFP_KERNEL);
			if (pinode == NULL) {
				PERROR("pvfs_revalidate_inode: out of memory!\n");
				unlock_kernel();
				read_unlock_inode(parent_pinode);
				PEXIT;
				return -ENOMEM;
			}

			*pinode = *old_pinode; /* copy in all the old values */
			init_pvfs_rwsem(&pinode->lock);
			pinode->name = (int8_t *)pinode + sizeof(struct pvfs_inode);
			inode->u.generic_ip = pinode;

			kfree(old_pinode);
		}
		
		/* copy the name into the inode: copy dirname, add /, copy name */
		ptr = pinode->name;
		strcpy(ptr, pvfs_inop(parent_inode)->name);
		ptr += len_dir;
		*ptr = '/';
		ptr++;
		strcpy(ptr, dentry->d_name.name);
	}

	PDEBUG(D_DIR, "pvfs_revalidate_inode called for %s (%ld).\n",
	pinode->name, (unsigned long) pinode->handle);

	/* Note: the only parameter that really is used inside of ll_pvfs_getmeta() 
	 *       is the file name. 
    */
	/* COHERENCE_DESIRED(getmeta) : Read lock on the entire file */
	error = ll_pvfs_getmeta(pinode, &meta, &phys);
	if (error < 0) {
		PDEBUG(D_DIR,
		"pvfs_revalidate: ll_pvfs_getmeta() failed; updating dcache\n");

		/* leave the dcache alone; the calling function handles cleanup */
		unlock_kernel();
		read_unlock_inode(parent_pinode);
		PEXIT;
		return error;
	}

	/* NOTE: not doing anything with the phys for now */
	pvfs_meta_to_inode(&meta, inode);

	if (old_handle != meta.handle) {
		/* need to update the inode and redo the hash */
		PDEBUG(D_DIR, 
				 "pvfs_revalidate_inode: handle changed from %ld to %ld for %s\n",
				 (unsigned long) old_handle, 
				 (unsigned long) meta.handle, pinode->name);

		remove_inode_hash(inode);
		inode->i_ino = meta.handle;
		pinode->handle = meta.handle;
		insert_inode_hash(inode);
	}
	if (old_mode != inode->i_mode) {
		/* need to update function pointers */
		if (S_ISDIR(inode->i_mode)) {
			inode->i_op = &pvfs_dir_inode_operations;
			inode->i_fop = &pvfs_dir_operations;
		}
		else if (S_ISREG(inode->i_mode)) {
			inode->i_op = &pvfs_file_inode_operations;
			inode->i_fop = &pvfs_file_operations;
			inode->i_data.a_ops = &pvfs_file_aops;
		}
		else {
			PERROR("pvfs_revalidate_inode: handle refers to something other than a file or a directory?\n");
			error = -EINVAL;
		}
	}

	unlock_kernel();
	read_unlock_inode(parent_pinode);
	PEXIT;
	return error;
}

/* pvfs_dentry_revalidate()
 *
 * A return value of "1" indicates that things are ok.  A return value of "0"
 * should result in the upper layers trying to prune out the dentry (bad).
 */
static int pvfs_dentry_revalidate(struct dentry *dentry, int flags)
{
	int ret;
	PENTRY;

	PDEBUG(D_DIR, "pvfs_dentry_revalidate called for %s.\n",
	dentry->d_name.name);

	/* don't call revalidate inode if the inode isn't valid */
	if (dentry->d_inode == NULL) {
		PDEBUG(D_DIR, "pvfs_dentry_revalidate: NULL inode (negative dentry) hit.\n");
		return 0;
		PEXIT;
	}

	ret = pvfs_revalidate_inode(dentry);
	PEXIT;
	if (ret == 0) return 1;
	else return 0; /* we got an error - let it be handled elsewhere. */
}


/* pvfs_meta_to_inode()
 *
 * Copies a pvfs_meta structure's values into an inode structure.
 */
int pvfs_meta_to_inode(struct pvfs_meta *mbuf, struct inode *ibuf)
{
	int error = 0;
	PENTRY;

	/* leaving the handle/i_ino fields alone for now */

	ibuf->i_mode = mbuf->mode;
	ibuf->i_uid = mbuf->uid;
	ibuf->i_gid = mbuf->gid;
	ibuf->i_size = mbuf->size;
	ibuf->i_atime = mbuf->atime;
	ibuf->i_mtime = mbuf->mtime;
	ibuf->i_ctime = mbuf->ctime;
	ibuf->i_blksize = mbuf->blksize;
	ibuf->i_blocks = mbuf->blocks;

	PEXIT;
	return error;
}

/*
 * Local variables:
 *  c-indent-level: 3
 *  c-basic-offset: 3
 *  tab-width: 3
 *
 * vim: ts=3
 * End:
 */ 
