/*
 * copyright (c) 1999 Rob Ross and Phil Carns, all rights reserved.
 *
 * Written by Rob Ross and Phil Carns, funded by Scyld Computing.
 *
 * This program is free software; you can redistribute it and/or
 * modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Contact:  Rob Ross    rbross@parl.clemson.edu
 *           Phil Carns  pcarns@parl.clemson.edu
 */

/* 
 * This code could potentially use 3 different methods for memory
 * transfers on read and write depending on which option is defined.
 * PVFS_BUFFER_DYNAMIC - dynamically allocate a buffer for each request
 * PVFS_BUFFER_STATIC - use a single statically allocated buffer
 * PVFS_BUFFER_MAPPED - use the kernel patch to map the user's buffer
 *   into kernel space, avoiding the copy and the allocate altogether.
 */

/* These calls are intended to handle communications between 
 *	the kernel low level vfs calls and the user level pvfsd daemon.
 */

#if !defined(__KPVFSD__)
#error "Cannot compile kpvfsdev.c without -D__KPVFSD__. Enable this before compiling."
#endif

#define __NO_VERSION__
#include "config.h"
#include "pvfs_kernel_config.h"
#include <linux/poll.h>
#ifdef HAVE_LINUX_IOBUF_H
#include <linux/iobuf.h>
#endif
#include "ll_pvfs.h"
#include "pvfs_bufmap.h"
#include "pvfs_linux.h"
#include "pvfsdev.h"
/* 2.4 */
#ifdef HAVE_LINUX_TQUEUE_H
#include <linux/tqueue.h>
#endif
#include <linux/slab.h>
#ifdef HAVE_LINUX_DEVFS_FS_KERNEL_H
#include <linux/devfs_fs_kernel.h>
#endif
#ifdef HAVE_LINUX_VMALLOC_H
#include <linux/vmalloc.h>
#endif

#include "kpvfsd.h"
#include "pvfs_proc.h"

/* and a maximum sequence number for kernel communications */
enum {
    PVFSDEV_MAX_SEQ = 8000
};


extern int pvfs_buffer;
extern int pvfs_maxsz;


/*********************************************************************
 * various structures used in the support code:
 */

/* buffer and size for memory operations */
struct pvfsdev_mem_reg {
    void *ptr;
    pvfs_size_t size;
};

/* this is the internal representation of the upcalls as needed to
 * be placed on a queue
 */
struct pvfsdev_upq_entry {
    struct pvfs_upcall pending_up;
    struct list_head req_list;
};

/* this is the internal representation of the downcalls as needed to
 * be placed on a queue
 */
struct pvfsdev_downq_entry {
    struct pvfs_downcall pending_down;
    struct list_head req_list;
};

/* sequence number entries in the seq_pending queue */
struct pvfsdev_seq_entry {
    pvfs_seq_t seq;
    struct list_head seq_list;
    int invalid;
    int kernel_read;		/* flag indicating that data is going to kernel space */

    /* not the most logical place to store this memory transfer
       * buffer information, but it will keep me from having to
       * manipulate an extra queue...
     */

    /* this is the kiobuf for reads and writes to user areas */
    struct kiobuf *iobuf;
    /* this is the pointer and offset for reads and writes to kernel
       * areas
     */
    struct pvfsdev_mem_reg kernelbuf;
};




/*********************************************************************
 * internal support functions:
 */

/* static int queue_len(struct list_head *my_list); */
static int down_dequeue(struct list_head *my_list,
			pvfs_seq_t my_seq,
			struct pvfs_downcall *out_downcall);
static int seq_is_pending(struct list_head *my_list, pvfs_seq_t my_seq,
			  int *invalid);
static int pvfsdev_down_enqueue(struct pvfs_downcall *in_downcall);
static int remove_seq_pending(struct list_head *my_list,
			      pvfs_seq_t my_seq);
static int invalidate_seq_pending(struct list_head *my_list,
				  pvfs_seq_t my_seq);
static pvfs_seq_t inc_locked_seq(void);
static int up_dequeue(struct list_head *my_list, pvfs_seq_t my_seq);
static struct pvfsdev_seq_entry *match_seq_entry(struct list_head *my_list,
						 pvfs_seq_t my_seq);
static int is_fake_useraddr(void);
static inline int pvfsdev_read_in_data(void *kptr, char *uptr,
				       pvfs_size_t size);
static inline int pvfsdev_write_out_data(char *uptr, void *kptr,
					 pvfs_size_t size);
static inline void cleanup_buffer(struct pvfs_upcall *in_upcall,
				  void **copybufp, void **pvfs_map_bufp);
static int setup_buffer(struct pvfs_upcall *in_upcall,
			void **copybufp, void **pvfs_map_bufp);
/**********************************************************************
 * various device internal queues and other device wide structures
 * that require locking before use. 
 */

/* this will be the queue of pending messages to pvfsd */
LIST_HEAD(up_pending);

/* this is the queue of upcall sequence numbers to be matched 
 * with the corresponding downcall in pvfsdev_enqueue.
 */
LIST_HEAD(seq_pending);

/* this is the list of pending downcalls */
LIST_HEAD(down_pending);

/* this the sequence number assigned to upcalls */
static pvfs_seq_t current_seq = 1;

/* this variable keeps up with the number of processes which have
 * opened the device at any given time.  Mostly for sanity checking
 * so that pvfs ops aren't attempted with no pvfsd.
 */
static int nr_daemons = 0;
atomic_t stop_new_requests; 

/* this is a buffer which will be used in transfers between the
 * application and pvfsd if static copies are used (the "static" buffer
 * option). Note that this variable now has external linkage as well. it is __not__ __static__ 
 * linkage anymore
 */

void *static_buf = NULL;

/* two semaphores:
 * static_buf_sem - protects our static buffer
 * seq_sem - protects sequence list
 */
static DECLARE_MUTEX(static_buf_sem);
static DECLARE_MUTEX(seq_sem);

/* keeps up with the amount of memory allocated for dynamic 
 * and mapped buffers 
 */
static unsigned long pvfsdev_buffer_size = 0;

/* these locks are used to protect the data defined above */
static spinlock_t pvfsdev_up_lock = SPIN_LOCK_UNLOCKED;
static spinlock_t pvfsdev_down_lock = SPIN_LOCK_UNLOCKED;
static spinlock_t pvfsdev_current_seq_lock = SPIN_LOCK_UNLOCKED;
static spinlock_t pvfsdev_nr_daemons_lock = SPIN_LOCK_UNLOCKED;
static spinlock_t pvfsdev_buffer_pool_lock = SPIN_LOCK_UNLOCKED;

/* three device wait queues:
 * pvfsdev_waitq - used for polling
 * pvfsdev_seq_wait - used for sleeping while waiting on a
 *   sequence-matched downcall
 * pvfsdev_buffer_pool_wait - used to sleep when dynamic buffer has
 *   gotten too large
 */
static DECLARE_WAIT_QUEUE_HEAD(pvfsdev_waitq);
static DECLARE_WAIT_QUEUE_HEAD(pvfsdev_seq_wait);
static DECLARE_WAIT_QUEUE_HEAD(pvfsdev_buffer_pool_wait);


/**********************************************************************
 * module specific code :
 */

/* pvfsdev_init()
 *
 * The kpvfsd daemon does not rely on the original character device interface 
 * and makes use of direct function calls to manipulate the device queues. 
 * This function simply sets up all the device queues.
 */

int pvfsdev_init(void)
{
	 atomic_set(&stop_new_requests,0);
    /* initialize the sequence counter */
    spin_lock(&pvfsdev_current_seq_lock);
    current_seq = 0;
    spin_unlock(&pvfsdev_current_seq_lock);
    spin_lock(&pvfsdev_nr_daemons_lock);
    nr_daemons = 0;
    spin_unlock(&pvfsdev_nr_daemons_lock);

    if (pvfs_buffer == PVFS_BUFFER_STATIC) {
			down(&static_buf_sem);
			static_buf = vmalloc(pvfs_maxsz);
			up(&static_buf_sem);
			if (static_buf == NULL)
				 return -ENOMEM;
    }
    PDEBUG(D_PSDEV, "pvfsdev: device queues initialized\n");
	 pvfs_sysctl_init();
    return 0;
}

/* pvfsdev_cleanup()
 *
 * Called at module unload.
 */
void pvfsdev_cleanup(void)
{
    struct pvfsdev_upq_entry *up_check;
    struct pvfsdev_downq_entry *down_check;
    struct pvfsdev_seq_entry *seq_check;
    int rm_seqs = 0;
    int inv_seqs = 0;

	 pvfs_sysctl_clean();

	 /* I have still not made use of this variable to stop new requests from
	  * hitting kpvfsd.. Need to fix that 
	  */
	 atomic_set(&stop_new_requests,1);

    /* we need to free up the all of the pending lists at this point */

    spin_lock(&pvfsdev_up_lock);
    while (up_pending.next != &up_pending) {
			up_check = list_entry(up_pending.next, struct pvfsdev_upq_entry, req_list);
			list_del(up_pending.next);
			kfree(up_check);
    }
    spin_unlock(&pvfsdev_up_lock);

    spin_lock(&pvfsdev_down_lock);
    while (down_pending.next != &down_pending) {
			down_check = list_entry(down_pending.next, struct pvfsdev_downq_entry, req_list);
			list_del(down_pending.next);
			kfree(down_check);
    }
    spin_unlock(&pvfsdev_down_lock);

    down(&seq_sem);
    while (seq_pending.next != &seq_pending) {
			seq_check = list_entry(seq_pending.next, struct pvfsdev_seq_entry, seq_list);
			rm_seqs++;
			if (seq_check->invalid) {
				 inv_seqs++;
			}
			list_del(seq_pending.next);
			kfree(seq_check);
    }
    up(&seq_sem);

    if (rm_seqs) {
			PERROR ("pvfsdev: %d upcall sequences removed, %d of which had become invalid.\n",
					  rm_seqs, inv_seqs);
    }

	 /* get rid of the static buffer we had laying around */
    if (pvfs_buffer == PVFS_BUFFER_STATIC) {
			down(&static_buf_sem);
			if (static_buf != NULL) {
				 vfree(static_buf);
				 static_buf = NULL;
			}
			up(&static_buf_sem);
    }
    PDEBUG(D_PSDEV, "pvfsdev: device queues cleaned up\n");
    return;
}


/***************************************************************
 * file operation code:
 */

/* pvfsdev_open()
 *
 * kpvfsd: calls this function to set things up.
 * All that this function does it to increase the number of 
 * daemons which have opened this "pseudo-device".
 * Provides a way by which we can keep track and/or respawn
 * the "kpvfsd" thread.
 *
 * Returns 0 on success, error number on failure.
 */
int pvfsdev_open(void)
{
    spin_lock(&pvfsdev_nr_daemons_lock);
    /* we really only allow one daemon to use this device at a time .
	 	 If we did want more than one kpvfsd threads, then one should be the
		 master and the rest should be slave threads feeding off the master
	  */
    if (nr_daemons != 0) {
			spin_unlock(&pvfsdev_nr_daemons_lock);
			return (-EBUSY);
    }
    nr_daemons++;
    spin_unlock(&pvfsdev_nr_daemons_lock);
	 PDEBUG(D_PSDEV," Opened device \n");
    return (0);
}

/* pvfsdev_release()
 *
 * kpvfsd: calls this function to indicate closing of "pseudo-device".
 *
 */
int pvfsdev_release(void)
{
    spin_lock(&pvfsdev_nr_daemons_lock);
    nr_daemons--;
    spin_unlock(&pvfsdev_nr_daemons_lock);
	 PDEBUG(D_PSDEV," Closed device \n");
    return (0);
}

static inline void kpvfsd_cleanup_buffer(struct pvfs_downcall *in_downcall)
{
	if(pvfs_buffer == PVFS_BUFFER_DYNAMIC) {
		/* check if it is a read/write operation */
		if(in_downcall->type == READ_OP || in_downcall->type == WRITE_OP) {
			/* only if we have allocated a temporary buffer, do we deallocate memory */
			if(in_downcall->xfer.from_kmem == 0) {
				if(in_downcall->xfer.ptr) {
					PDEBUG(D_PSDEV, "kpvfsd freeing %p size %ld\n",in_downcall->xfer.ptr, (unsigned long)in_downcall->xfer.orig_size); 
					vfree(in_downcall->xfer.ptr);
				}
				spin_lock(&pvfsdev_buffer_pool_lock);
				pvfsdev_buffer_size -= in_downcall->xfer.orig_size;
				spin_unlock(&pvfsdev_buffer_pool_lock);
				wake_up_interruptible(&pvfsdev_buffer_pool_wait);
			}
		}
	}
	return;
}


/* pvfsdev_write()
 *
 * kpvfsd: calls this function to write a downcall into the "pseudo-device"
 *
 * Returns 0 on success, error number on failure.
 */
ssize_t pvfsdev_write(struct pvfs_downcall *in_req, size_t buf_sz, long timeout)
{
    int req_sz = sizeof(struct pvfs_downcall);
    int ret = 0;
    int invalid = 0;
    struct pvfsdev_seq_entry *my_seq = NULL;

    if (buf_sz != req_sz || !in_req) {
		 	PDEBUG(D_DOWNCALL, "Invalid parameters (write) %d,%d,%p\n",buf_sz,req_sz,in_req);
			return -EINVAL;
    }

    /* verify the magic */
    if (in_req->magic != PVFS_DOWNCALL_MAGIC) {
		 	PDEBUG(D_DOWNCALL," Invalid downcall magic number\n");
			return -EINVAL;
    }

    if (in_req->seq == -1) {
		 	PDEBUG(D_DOWNCALL," Independant downcall sequence number\n");
			/* this is an independent downcall, do not match it to
			 * an outstanding upcall sequence. 
			 */
			/* do something else instead- not defined yet. */
			return -ENOSYS;
    }

    PDEBUG(D_TIMING, "pvfsdev: writing downcall number %d.\n", in_req->seq);

    /* ok- we are going to grab this semaphore right here, and then hang
       * on to it for a long time to make sure that no one invalidates our
       * sequence number while we are completing this downcall!
     */
    down(&seq_sem);

    /* check to see if we are expecting this sequence number */
    ret = seq_is_pending(&seq_pending, in_req->seq, &invalid);
    if (!ret) {
			/* Uh-oh... something is amok because we don't have the
			 * sequence number for this downcall queued up... bail
			 * out! 
			 */
			up(&seq_sem);
			PDEBUG(D_TIMING, "pvfsdev: can't match this sequence number.\n");
			return -EINVAL;
    }

    /* at this point we need to be sure this write is still valid
       * with respect to its comm sequence.  If not just eat it and
       * return as normal- the user app on the the other side was
       * probably terminated. 
     */
    if (invalid) {
			/* what sort of op has been invalidated? */
			my_seq = match_seq_entry(&seq_pending, in_req->seq);
			/* 
				Only the user-space pvfsd design has the 2nd stage for the 
				write operation, the kpvfsd design does not have such a phase,
				So we can go ahead and remove the pending sequence element.
			*/
			PDEBUG(D_TIMING, "pvfsdev: received write for invalid seq.\n");
			ret = remove_seq_pending(&seq_pending, in_req->seq);
			up(&seq_sem);
			if (ret < 0) {
				 PDEBUG(D_TIMING, "pvfsdev: remove of seq_entry failed... \n");
			}
			/* cleanup the buffer */
			kpvfsd_cleanup_buffer(in_req);
			return req_sz;
    }

    /* now everything should be Ok for submitting this downcall */

    /* use the switch here to catch special test cases as well as
     * normal requests that require memory transfers. 
     */

	 /* Reads and writes need separate memory transfers only if pvfsd lives in a separate
	    address space from the device driver interface. For the kpvfsd case, this is not 
		 the case and hence reads and writes do not incur extra memory xfers . They can be staged
		 off the kernel allocated memory
	 */
    ret = pvfsdev_down_enqueue(in_req);
    if (ret < 0) {
			up(&seq_sem);
			return ret;
    }
    up(&seq_sem);
    return req_sz;
}


/* pvfsdev_read_in_data()
 *
 * Parameters:
 * kptr - kernel space destination, could be an address or a kiobuf
 * uptr - user space (source) address pointer
 * size - size of data to move
 */
static inline int pvfsdev_read_in_data(void *kptr, char *uptr, pvfs_size_t size)
{
    int ret;
#ifdef HAVE_PVFS_BUFMAP
    if (pvfs_buffer == PVFS_BUFFER_MAPPED) {
			/* in this case we are copying from one userspace to another */
			ret = pvfs_copy_to_userbuf((struct kiobuf *) kptr, uptr, size);
			return ret;
    }
#endif
    ret = copy_from_user((char *) kptr, uptr, size);
    return ret;
}

/* pvfsdev_write_out_data()
 *
 * Parameters:
 * uptr - user space (destination) address pointer
 * kptr - kernel space source, could be an address or a kiobuf
 * size - size of data to move
 */
static inline int pvfsdev_write_out_data(char *uptr, void *kptr, pvfs_size_t size)
{
    int ret;
#ifdef HAVE_PVFS_BUFMAP
    if (pvfs_buffer == PVFS_BUFFER_MAPPED) {
			/* in this case we are copying from one userspace to another */
			ret = pvfs_copy_from_userbuf(uptr, (struct kiobuf *) kptr, size);
			return ret;
    }
#endif
    ret = copy_to_user(uptr, (char *) kptr, size);
    return ret;
}


/* pvfsdev_read()
 *
 * Called by kpvfsd as a direct function call.
 *
 * Returns 0 on success, error number on failure. 
 * this function can __block__
 */
ssize_t pvfsdev_read(struct pvfs_upcall *buf,size_t buf_sz, long timeout)
{

    struct pvfsdev_upq_entry *out_entry = NULL;
    int retval = 0;
    DECLARE_WAITQUEUE(wait, current);

    if (buf_sz != sizeof(struct pvfs_upcall) || !buf) {
		 	PDEBUG(D_UPCALL, "Invalid parameter %d,%d,%p\n",buf_sz,sizeof(struct pvfs_upcall),buf);
			return -EINVAL;
    }

    /* put ourselves on a wait queue while we check to see if there is
       * really an upcall available to read.  If it isn't there, we will
       * schedule ourselves to wait on it.
     */
	 retval = 0;
    wait.task = current;
    add_wait_queue(&pvfsdev_waitq, &wait);
    set_current_state(TASK_INTERRUPTIBLE);
    while (up_pending.prev == &up_pending) {
			if (signal_pending(current)) {
				 retval = -ERESTARTSYS;
				 break;
			}
			schedule();
    }

    remove_wait_queue(&pvfsdev_waitq, &wait);
    set_current_state(TASK_RUNNING);

    if (retval) {
			/* we were interrupted or are in non-blocking mode w/ nothing to do */
			return (retval);
    }

    /* if we reach this point in the code, there should be an upcall to
       * read, unless the other side somehow manages to abort before we
       * actually read it out
     */

    /* get the upcall off of the queue and copy it to kpvfsd */
    spin_lock(&pvfsdev_up_lock);
    if (up_pending.prev != &up_pending) {
			out_entry = list_entry(up_pending.prev, struct pvfsdev_upq_entry, req_list);
			PDEBUG(D_TIMING, "pvfsdev: reading upcall number %d.\n", out_entry->pending_up.seq);
			list_del(up_pending.prev);
			spin_unlock(&pvfsdev_up_lock);
			PDEBUG(D_ALLOC,"Copying from %p to %p\n",&(out_entry->pending_up),buf);
			memcpy(buf, &(out_entry->pending_up), sizeof(struct pvfs_upcall));
			kfree(out_entry);
			return (sizeof(struct pvfs_upcall));
    }
	 else {
			/* nothing in the queue */
			spin_unlock(&pvfsdev_up_lock);
			PDEBUG(D_PSDEV,"Nothing in the queue\n");
			return 0; //will make kpvfsd() close some files and then re-enter this routine
    }
}

/*******************************************************************
 * support functions:
 */

/* pvfsdev_enqueue()
 *
 * this adds upcalls to the outgoing queue.  it will then wait on a matching
 * downcall before returning.
 *
 * TODO: Optimize better for the no-patch case when we feel like it.
 * Returns 0 on success, error  number on failure.
 * 
 */
int pvfsdev_enqueue(struct pvfs_upcall *in_upcall, struct pvfs_downcall *out_downcall)
{

    struct pvfsdev_upq_entry *q_upcall = NULL;
    struct pvfsdev_seq_entry *q_seq = NULL;
    struct pvfsdev_seq_entry *q_seq_probe = NULL;
    DECLARE_WAITQUEUE(my_wait, current);
    struct task_struct *tsk = current;
    pvfs_seq_t my_seq = 0;
    int err = 0;
    int mem_xfer = 0;
    void *copybuf = NULL;
    int mem_read = 0, no_memory_xfers = 1;
	 struct timeval begin = {0,0}, end = {0, 0};


    /* pvfs_map_buf and rw are only used in the PVFS_BUFFER_MAPPED case */
    void *pvfs_map_buf = NULL;
    int rw = -1;
	 pvfs_callstat.total++;

    /* nr_daemons tells us if there is a user-space pvfsd listening for 
       * upcalls.  If there isn't, we bail out.
     */
    if (nr_daemons < 1) {
			int ret = -ENODEV;
			/* we can try to respawn the kpvfsd sub-system since it 
				must have exited if this condition were to hold true */
			ret = init_kpvfsd();	
			if(ret) {
					PERROR("pvfsdev : Could not respawn the kpvfsd sub-system %d\n",ret);
					return ret;
			}
			PDEBUG(D_PSDEV,"Re-spawned the kpvfsd sub-system\n");
    }
	 if (atomic_read(&stop_new_requests) != 0) {
		 	PDEBUG(D_PSDEV,"Stop new requests\n");
	 		return -EAGAIN;
	 }

    q_seq = kmalloc(sizeof(struct pvfsdev_seq_entry), GFP_KERNEL);
    if (!q_seq) {
			return -ENOMEM;
    }
    q_seq->invalid = 0;

    /* here we are going to check to see if the upcall involves a
       * memory transfer- if so we are going to go ahead and set up the
       * pvfs io mapping, if needed.
     */
    switch (in_upcall->type) {
    case READ_OP:
    case WRITE_OP:
	 {
			no_memory_xfers = 0;
			if ((in_upcall->xfer.to_kmem != 0) || (is_fake_useraddr())) {
				 /* not a user-space buffer; don't map */
				 q_seq->kernelbuf.ptr = in_upcall->xfer.ptr;
				 q_seq->kernelbuf.size = in_upcall->xfer.size;
				 q_seq->kernel_read = 1;
				 break;
			}
			/* wait on buffer resources and allocate them.  Mapping of user space
			 * buffers into kernel space (in the PVFS_BUFFER_MAPPED case) is handled
			 * a little further down 
			 */
			err = setup_buffer(in_upcall, &copybuf, &pvfs_map_buf);
			if (err)
				 return err;

			/* buffers are all set; let's go. */
			if (in_upcall->type == WRITE_OP) {
				 /* file system write operation */
				 if (pvfs_buffer == PVFS_BUFFER_MAPPED) {
						rw = PVFS_BUF_READ;	/* only used by the mapped technique */
				 } 
				 else {
				 	   /* copy the data to be written from user to device/kernel memory */
						err = copy_from_user(copybuf, in_upcall->xfer.ptr, in_upcall->xfer.size);
						if (err) {
							 PERROR ("pvfsdev: pvfs memory transfer setup failure. %d\n",err);
							 cleanup_buffer(in_upcall, &copybuf, &pvfs_map_buf);
							 return err;
						}
				 }
			} 
			else {
				 /* file system read operation */
				 if (pvfs_buffer == PVFS_BUFFER_MAPPED) {
						rw = PVFS_BUF_WRITE;
				 } 
				 else {
						/* this is a read xfer we need to handle later */
						mem_read = 1;
				 }
			}

			if (pvfs_buffer == PVFS_BUFFER_MAPPED) {
#ifdef HAVE_PVFS_BUFMAP
				 /* map the user buffer into kernel space and point the queue entry
				  * to the right buffer
				  */
				 err = pvfs_map_userbuf(rw, in_upcall->xfer.ptr, in_upcall->xfer.size, &pvfs_map_buf);
				 if (err) {
						PERROR("pvfsdev: pvfs memory transfer setup failure.%d\n",err);
						cleanup_buffer(in_upcall, &copybuf, &pvfs_map_buf);
						return err;
				 }
				 q_seq->iobuf = pvfs_map_buf;
#endif
			}
			else {
				 /* point the queue entry to the right buffer */
				 q_seq->iobuf = copybuf;
			}
			q_seq->kernel_read = 0;
			mem_xfer = 1;
			break;
	 }
	 case GETDENTS_OP:
	 {
			no_memory_xfers = 0;
			if (in_upcall->xfer.to_kmem == 0) {
				 /* not supported at this time */
				 PDEBUG(D_SPECIAL, "pvfsdev: user-space address to GETDENTS?\n");
				 return -EINVAL;
			}
			q_seq->kernelbuf.ptr = in_upcall->xfer.ptr;
			q_seq->kernelbuf.size = in_upcall->xfer.size;
			q_seq->kernel_read = 1;
			break;
	 }
    default:
			/* not a memory transfer operation */
			break;
    }
    /* creating the structure to put on the upcall queue */
    q_upcall = kmalloc(sizeof(struct pvfsdev_upq_entry), GFP_KERNEL);
    if (!q_upcall) {
			/* FIXME: Is q_seq a memory leak here? */
			return -ENOMEM;
    }
	 PDEBUG(D_ALLOC,"Copying from %p to %p\n",in_upcall,&q_upcall->pending_up);
	 memcpy(&q_upcall->pending_up,in_upcall,sizeof(struct pvfs_upcall));

	 /* Change some of the upcall fields, i.e pass initialized
	 	 pointers so that kpvfsd has to just stage the operation
		 in the passed pointer only if it is a memory xfer.
		 for mmap,GETDENTS : q_seq->kernelbuf.ptr seems to be the right value
		 for write,read-static,read-dynamic,read-mapped,write-mapped q_seq->iobuf seems
		 to be the right value
		 In the modified upcall, the field to_kmem indicates whether or not the address
		 is a userbuf or not. I am not sure if that last case will work.
		 Need to better understand the memory-mapping.
	  */
	 if(no_memory_xfers == 0) {
				q_upcall->pending_up.xfer.to_kmem = 1; 
				if(in_upcall->xfer.to_kmem != 0) {
						q_upcall->pending_up.xfer.ptr = q_seq->kernelbuf.ptr;
				}
				else {
						q_upcall->pending_up.xfer.ptr = q_seq->iobuf ;
						if(pvfs_buffer == PVFS_BUFFER_MAPPED) {
#ifdef HAVE_PVFS_BUFMAP
							  q_upcall->pending_up.xfer.to_kmem = 0;
#endif
						}
				}
				PDEBUG(D_ALLOC,"pvfsdev: Initialized ptr to %p \n",q_upcall->pending_up.xfer.ptr);
	 }
    /* grab a sequence number atomically */
    my_seq = inc_locked_seq();
    q_upcall->pending_up.seq = my_seq;
    q_seq->seq = my_seq;

    /* 
	  * atomically modify the sequence number list.  this is done to let us keep
	  * sequence numbers around in case an upcall is aborted for some reason 
	  * (there might be some other reason also???)
     */
    down(&seq_sem);
    list_add(&(q_seq->seq_list), &seq_pending);
    up(&seq_sem);

    /* atomically add ourselves to the upcall queue */
    spin_lock(&pvfsdev_up_lock);
    list_add(&(q_upcall->req_list), &up_pending);
    spin_unlock(&pvfsdev_up_lock);

    /* wake up processes listening on pvfsdev for requests (pvfsd) */
    wake_up_interruptible(&pvfsdev_waitq);

    /* now I want to wait here until a downcall comes in 
       * that matches the sequence number.  i stay on the waitqueue until 
       * I explicitly remove myself.
     */
    my_wait.task = tsk;
    add_wait_queue(&pvfsdev_seq_wait, &my_wait);
	 /* this variable is set thru the sysctl interface */
	 if(pvfs_upcall_timestamping) {
		 do_gettimeofday(&begin);
	 }
  seq_repeat:
    set_current_state(TASK_INTERRUPTIBLE);

    /* this lets this mechanism be interrupted if needed */
    /* but don't interrupt a hint operation!  This is necessary
       * to correctly close sockets and such on program exit.
     */
    /* check for bad signals.  if we got one, bail out gracefully */
    if ((check_for_signal_to_kpvfsd())) {
			PDEBUG(D_PSDEV, "Process %d inside pvfsdev: recieved signal- bailing out.\n",current->pid);
			/* need to do something reasonable here... */
			/* so subsequent calls don't freak out :) */

			/* at least need to dequeue above stuff and make sure the
			 * module use count is correct
			 */
			remove_wait_queue(&pvfsdev_seq_wait, &my_wait);
			set_current_state(TASK_RUNNING);

			/* how far along did the kpvfsd get on this operation?  Three
			 * scenarios could occur:
			 *
			 * 1) The kpvfsd has not read the upcall yet.  We need to dequeue
			 * the upcall and then dequeue and destroy the sequence entry that
			 * matches it.
			 *
			 * 2) The kpvfsd has read the upcall, but has not yet produced a
			 * downcall in response.  That means  that we should invalidate 
			 * the sequence number.  The pvfsdev_write function
			 * will detect that the sequence was invalidated and silently eat
			 * the matching downcall. Special case here since user program cannot
			 * cleanup the buffer, but let kpvfsd deal with freeing the buffer.
			 *
			 * 3) The pvfsd has read the upcall, and has already completed the
			 * downcall response. In this case, we need to get the
			 * downcall out of the queue and destroy it before we return.
			 */

			/* see if we can find the upcall and remove it */
			err = up_dequeue(&up_pending, my_seq);
			if (err == 0) {
				 /* the upcall has not been read */
				 /* scenario 1. */
				 down(&seq_sem);
				 remove_seq_pending(&seq_pending, my_seq);
				 up(&seq_sem);
				 if (mem_xfer)
					 cleanup_buffer(in_upcall, &copybuf, &pvfs_map_buf);
			}
			else {
				 /* the upcall has been read */
				 down(&seq_sem);
				 /* has the downcall completed? */
				 q_seq_probe = match_seq_entry(&seq_pending, my_seq);
				 if (q_seq_probe) {
						/* downcall not completed */
						/* scenario 2 */
						invalidate_seq_pending(&seq_pending, my_seq);
						/* we should not let user call cleanup buffer in this case */
				 } 
				 else {
						/* downcall has completed */
						/* scenario 3 */
						down_dequeue(&down_pending, my_seq, out_downcall);
						if (mem_xfer)
						 	cleanup_buffer(in_upcall, &copybuf, &pvfs_map_buf);
				 }
				 up(&seq_sem);
			}

			if(pvfs_upcall_timestamping) {
				do_gettimeofday(&end);
				 if (end.tv_usec < begin.tv_usec) {
					          end.tv_usec += 1000000; end.tv_sec--;
				 }
				 end.tv_sec  -= begin.tv_sec;
				 end.tv_usec -= begin.tv_usec;
				 /* update the upcall statistics */
				 pvfs_upcall_stats(in_upcall->type, (end.tv_sec * 1000000 + end.tv_usec));
			}
			return -EINTR;
    }

    /* here I need to check the seq_pending list and see if my sequence
       * number has been removed, indicating a matching downcall.  if not we call
       * schedule and go back to waiting.
     */
    down(&seq_sem);
    err = seq_is_pending(&seq_pending, my_seq, NULL);
    up(&seq_sem);
    if (err) {
			schedule();
			goto seq_repeat;
    }
    remove_wait_queue(&pvfsdev_seq_wait, &my_wait);
    set_current_state(TASK_RUNNING);
	if(pvfs_upcall_timestamping) {
		do_gettimeofday(&end);
		 if (end.tv_usec < begin.tv_usec) {
						 end.tv_usec += 1000000; end.tv_sec--;
		 }
		 end.tv_sec  -= begin.tv_sec;
		 end.tv_usec -= begin.tv_usec;
		 /* update the upcall statistics */
		 pvfs_upcall_stats(in_upcall->type, (end.tv_sec * 1000000 + end.tv_usec));
	 }

    /* now I need to pull my matched downcall off of the downcall
       * list and return
     */
    err = down_dequeue(&down_pending, my_seq, out_downcall);
    if (err < 0) {
			if (mem_xfer)
				 cleanup_buffer(in_upcall, &copybuf, &pvfs_map_buf);
			return err;
    }
    /* if the operation was a read, copy the data received from the daemon
	  * into the user buffer if necessary (i.e. not mapped)
     */
    if (mem_read && (pvfs_buffer == PVFS_BUFFER_DYNAMIC
		     || pvfs_buffer == PVFS_BUFFER_STATIC)) {
			PDEBUG(D_ALLOC, "Copying to user-space from %p to %p of size %ld\n",copybuf, in_upcall->xfer.ptr,(unsigned long) in_upcall->xfer.size);
			err = copy_to_user(in_upcall->xfer.ptr, copybuf, in_upcall->xfer.size);
    }

    /* if we did a transfer, clean up the buffer and wake people up */
    if (mem_xfer)
			cleanup_buffer(in_upcall, &copybuf, &pvfs_map_buf);
    return err;
}

/* setup_buffer()
 *
 * Used to allocate buffer space in the dynamic case, or to grab the
 * static_buf_sem semaphore in the static case.  We don't map the user buffer
 * for the mapped case yet, as there are some additional checks that must
 * first be made (SHOULD WE MOVE THE MAPPING INTO THIS FUNCTION?)
 *
 * The first half of this function handles checking resource limitations in
 * the mapped and dynamic case.  
 *
 * NOTE: pvfs_map_bufp is currently unused, as we're not doing the mapping in
 * here yet.
 */
static int setup_buffer(struct pvfs_upcall *in_upcall, void **copybufp, void **pvfs_map_bufp)
{
    struct task_struct *tsk = current;
    DECLARE_WAITQUEUE(my_wait, current);

    /* This following bit of code makes sure that the sum of 
       * all of the buffers used for the MAPPED or DYNAMIC buffers
       * doesn't exceed the maxsz defined at module load time.
     */
    if (pvfs_buffer == PVFS_BUFFER_DYNAMIC || pvfs_buffer == PVFS_BUFFER_MAPPED) {

			spin_lock(&pvfsdev_buffer_pool_lock);
			if ((pvfsdev_buffer_size + in_upcall->xfer.size) <= pvfs_maxsz) {
				 pvfsdev_buffer_size += in_upcall->xfer.size;
				 spin_unlock(&pvfsdev_buffer_pool_lock);
			} 
			else {
				 spin_unlock(&pvfsdev_buffer_pool_lock);
				 /* uh oh, we have exceeded the max buffer size available */
				 PDEBUG(D_SPECIAL, "exceeded buffer size: sleeping.\n");
				 my_wait.task = tsk;
				 add_wait_queue(&pvfsdev_buffer_pool_wait, &my_wait);
dynamic_repeat:
				 set_current_state(TASK_INTERRUPTIBLE);
				 /* check for bad signals and bail out if we got one */
				 if ((check_for_signal_to_kpvfsd())) {
						PDEBUG(D_PSDEV, "pvfsdev: recieved signal- bailing out.\n");
						remove_wait_queue(&pvfsdev_buffer_pool_wait, &my_wait);
						set_current_state(TASK_RUNNING);
						return -EINTR;
				 }

				 /* now we can look again to see if we can allocate a buffer */
				 spin_lock(&pvfsdev_buffer_pool_lock);
				 if ((pvfsdev_buffer_size + in_upcall->xfer.size) > pvfs_maxsz) {
						spin_unlock(&pvfsdev_buffer_pool_lock);
						schedule();
						/* go back to sleep if we can't */
						goto dynamic_repeat;
				 }

				 /* the memory is available- let's rock! */
				 pvfsdev_buffer_size += in_upcall->xfer.size;
				 spin_unlock(&pvfsdev_buffer_pool_lock);

				 remove_wait_queue(&pvfsdev_buffer_pool_wait, &my_wait);
				 set_current_state(TASK_RUNNING);
			}
    }

    /* Here we allocate our buffer for the dynamic copy, or simply
       * set the pointer to point to the static buffer in the static
       * case.  This value isn't used by the mapped buffer code.
     */
    if (pvfs_buffer == PVFS_BUFFER_DYNAMIC) {
			*copybufp = (void *) vmalloc(in_upcall->xfer.size);
			PDEBUG(D_PSDEV ,"process allocated %p size %ld\n",*copybufp, (unsigned long) in_upcall->xfer.size); 
			if (*copybufp == NULL) {
				 PERROR("pvfsdev: could not allocate memory!\n");
				 spin_lock(&pvfsdev_buffer_pool_lock);
				 pvfsdev_buffer_size -= in_upcall->xfer.size;
				 spin_unlock(&pvfsdev_buffer_pool_lock);
				 return -ENOMEM;
			}
    }
	 else if (pvfs_buffer == PVFS_BUFFER_STATIC) {
			/* go ahead and decrement the semaphore while we are at it */
			down(&static_buf_sem);
			*copybufp = static_buf;
    }
    /* this would be the place to map the user buffer if we wanted to
       * do it in this function.
     */
    return 0;
}


/* cleanup_buffer()
 *
 * Used to deallocate buffer space in the dynamic case, or to unmap user
 * buffers in the mapped case, or to release the static_buf_sem semaphore in
 * the static case.
 *
 * This function also handles waking up other processes waiting on these
 * resources.  Both the mapped and dynamic schemes use the
 * pvfsdev_buffer_pool_wait wait queue to wait on resources to be freed, so we
 * wake up the guys on that queue in here.  
 */
static inline void cleanup_buffer(struct pvfs_upcall *in_upcall,
				  void **copybufp, void **pvfs_map_bufp)
{
    if (pvfs_buffer == PVFS_BUFFER_MAPPED) {
#ifdef HAVE_PVFS_BUFMAP
			if (*pvfs_map_bufp) {
				 pvfs_unmap_userbuf(*pvfs_map_bufp);
				 *pvfs_map_bufp = NULL;
			}
			spin_lock(&pvfsdev_buffer_pool_lock);
			pvfsdev_buffer_size -= in_upcall->xfer.size;
			spin_unlock(&pvfsdev_buffer_pool_lock);
			wake_up_interruptible(&pvfsdev_buffer_pool_wait);
#endif
    }
	 else if (pvfs_buffer == PVFS_BUFFER_DYNAMIC) {
			if (*copybufp) {
				 PDEBUG(D_PSDEV,"process freeing %p size %ld\n",*copybufp, (unsigned long)in_upcall->xfer.size);   
				 vfree(*copybufp);
				 *copybufp = NULL;
			}
			spin_lock(&pvfsdev_buffer_pool_lock);
			pvfsdev_buffer_size -= in_upcall->xfer.size;
			spin_unlock(&pvfsdev_buffer_pool_lock);
			wake_up_interruptible(&pvfsdev_buffer_pool_wait);
    }
	 else if (pvfs_buffer == PVFS_BUFFER_STATIC) {
			up(&static_buf_sem);
    }
    return;
}

/* check_for_signal_to_kpvfsd()
 *
 * This function checks for pending signals after a process has woken up.
 *
 * Returns 0 if everything is ok, -EINTR otherwise.
 */
int check_for_signal_to_kpvfsd(void)
{
    if (signal_pending(current)) {
			if (sigismember(&(current->pending.signal), SIGKILL)
				 || sigismember(&(current->pending.signal), SIGTERM)
				 || sigismember(&(current->pending.signal), SIGINT))
			{
				 return -EINTR;
			}
    }
    return 0;
}


/* pvfsdev_enqueue_noresponse()
 *
 * This function is used to queue an upcall for which there will be no
 * matching downcall.  This is indicated to the daemon by a -1 sequence
 * number, which is stored in the queue structure (q_upcall). 
 */
int pvfsdev_enqueue_noresponse(struct pvfs_upcall *in_upcall)
{
    struct pvfsdev_upq_entry *q_upcall = NULL;

    /* nr_daemons tells us if there is a user-space pvfsd or kpvfsd listening 
	 	for upcalls.  If there isn't, we bail out in the case of pvfsd or respawn 
		a new kpvfsd 
     */
    if (nr_daemons < 1) {
	 		int ret = -ENODEV;
			/* we can try to respawn the kpvfsd sub-system since it 
				must have exited if this condition were to hold true */
			ret = init_kpvfsd();	
			if(ret) {
					PERROR("pvfsdev : Could not respawn the kpvfsd sub-system %d\n",ret);
					return ret;
			}
    }
	 if (atomic_read(&stop_new_requests) != 0) {
	 		return -EAGAIN;
	 }
    q_upcall = kmalloc(sizeof(struct pvfsdev_upq_entry), GFP_KERNEL);
    if (!q_upcall) {
			return -ENOMEM;
    }

    /* this q_upcall structure is the element that sits in the queue */
	 PDEBUG(D_ALLOC,"Copying from %p to %p\n",&q_upcall->pending_up,in_upcall);
	 memcpy(&q_upcall->pending_up,in_upcall,sizeof(struct pvfs_upcall));
    q_upcall->pending_up.seq = -1;

    /* add this upcall to the queue */
    spin_lock(&pvfsdev_up_lock);
    list_add(&(q_upcall->req_list), &up_pending);
    spin_unlock(&pvfsdev_up_lock);

    /* wake up processes reading or selecting on our device */
    wake_up_interruptible(&pvfsdev_waitq);

    return 0;
}


/* seq_is_pending()
 *
 * use this to find out if a sequence number is in the seq list
 *
 * now the invalid integer will be set to alert the caller that they
 * have matched a sequence number that has been declared invalid.
 * if they caller does not want to know about that they should set
 * that pointer to NULL
 *
 * IMPORTANT: this function assumes that the caller holds a semaphore
 * protecting the sequence list!
 */
static int seq_is_pending(struct list_head *my_list, pvfs_seq_t my_seq,
			  int *invalid)
{
    struct list_head *check = NULL;
    struct pvfsdev_seq_entry *entry = NULL;
    for (check = my_list->next; check != my_list; check = check->next) {
			entry = list_entry(check, struct pvfsdev_seq_entry, seq_list);
			if (entry->seq == my_seq) {
				 if (invalid != NULL) {
						*invalid = entry->invalid;
				 }
				 return 1;
			}
    }
    return 0;
}


/* down_dequeue()
 *
 * use this to figure out which downcall to dequeue based on which
 * one matches your sequence number-return null on error...
 *
 * now also returns the fs from the downcall side :)
 *
 * Returns 0 on success,error number on failure.
 */
static int down_dequeue(struct list_head *my_list, pvfs_seq_t my_seq,
			struct pvfs_downcall *out_downcall)
{
    struct list_head *check = NULL;
    struct pvfsdev_downq_entry *entry = NULL;


    /* iterate through list of entries looking for sequence number;
       * we expect it to be near the head of the list 
     */
    spin_lock(&pvfsdev_down_lock);
    for (check = my_list->next; check != my_list; check = check->next) {
			entry = list_entry(check, struct pvfsdev_downq_entry, req_list);
			if (entry->pending_down.seq == my_seq) {
				 /* found it.  return copy of downcall, remove entry from list, 
				  * free downcall entry structure, unlock list, and return.
				  */
				 PDEBUG(D_ALLOC, "Copied from %p to %p\n",&entry->pending_down,out_downcall);
				 memcpy(out_downcall,&entry->pending_down,sizeof(struct pvfs_downcall));
				 list_del(check);
				 kfree(entry);
				 /* we don't have to kfree() pending_down because it is a part of the
				  * pvfsdev_downq_entry structure
				  */
				 spin_unlock(&pvfsdev_down_lock);
				 return 0;
			}
    }
    spin_unlock(&pvfsdev_down_lock);
    /* looked through the whole list and didn't find a match */
    PDEBUG(D_DOWNCALL, "pvfsdev: didn't find matching downcall in queue!\n");
    return -EINVAL;
}


/* pvfsdev_down_enqueue()
 *
 * this adds downcalls to the incoming queue
 *
 * IMPORTANT!  This function assumes that the caller holds a semaphore
 * protecting the sequence list before it is called!
 *
 * Returns 0 on success, error number on failure.
 */
static int pvfsdev_down_enqueue(struct pvfs_downcall *in_downcall)
{
    struct pvfsdev_downq_entry *q_downcall;
    int ret = -1;

    /* we also need to handle removing the sequence number from the
       * list of pending sequence numbers for upcall/downcall pairs
     */
    q_downcall = kmalloc(sizeof(struct pvfsdev_downq_entry), GFP_KERNEL);
    if (!q_downcall) {
			return -ENOMEM;
    }
    /* make a copy of the downcall */
	 PDEBUG(D_ALLOC,"Copying downcall from %p to %p\n",in_downcall,&q_downcall->pending_down);
    memcpy(&q_downcall->pending_down,in_downcall,sizeof(struct pvfs_downcall));
    spin_lock(&pvfsdev_down_lock);
    list_add(&(q_downcall->req_list), &down_pending);
    spin_unlock(&pvfsdev_down_lock);
    /* take the sequence number out of the sequence number list */
    ret = remove_seq_pending(&seq_pending, in_downcall->seq);
    if (ret < 0) {
			PDEBUG(D_DOWNCALL, "pvfsdev: remove of seq_entry failed... \n");
    }
    /* wake stuff up - see poll function */
    wake_up_interruptible(&pvfsdev_seq_wait);
    return 0;
}


/* remove_seq_pending()
 *
 * use this to remove a sequence number from the pending sequence
 * list
 *
 * IMPORTANT!  This function assumes that the caller already holds a
 * semaphore to protect the sequence list!
 *
 * Returns 0 on success, -errno on failure.
 */
static int remove_seq_pending(struct list_head *my_list, pvfs_seq_t my_seq)
{
    struct list_head *check = NULL;
    struct pvfsdev_seq_entry *entry = NULL;

    for (check = my_list->next; check != my_list; check = check->next) {
			entry = list_entry(check, struct pvfsdev_seq_entry, seq_list);
			if (entry->seq == my_seq) {
				 list_del(check);
				 kfree(entry);
				 return 0;
			}
    }
    return -EINVAL;
}


/* inc_locked_seq()
 *
 * this simply increments the communication sequence number and
 * returns its new value.
 */
static pvfs_seq_t inc_locked_seq()
{
    pvfs_seq_t ret = 0;
    spin_lock(&pvfsdev_current_seq_lock);
    if (current_seq < PVFSDEV_MAX_SEQ) {
			current_seq++;
    } 
	 else {
			current_seq = 1;
    }
    ret = current_seq;
    spin_unlock(&pvfsdev_current_seq_lock);
    return ret;
}


/* up_dequeue()
 *
 * this function dequeues specific pending upcalls- mostly used for
 * cleanup mechanisms
 *
 * Returns 0 on success, -errno on failure.
 */
static int up_dequeue(struct list_head *my_list, pvfs_seq_t my_seq)
{
    struct list_head *check = NULL;
    struct pvfsdev_upq_entry *entry = NULL;

    /* iterate through list of entries looking for matching sequence number.
       * we go through the list from oldest to newest, because we expect we're 
       * looking for an old entry.
     */
    spin_lock(&pvfsdev_up_lock);
    for (check = up_pending.prev; check != &up_pending; check = check->prev) {
			entry = list_entry(up_pending.prev, struct pvfsdev_upq_entry, req_list);
			if (entry->pending_up.seq == my_seq) {
				 list_del(up_pending.prev);
				 kfree(entry);
				 spin_unlock(&pvfsdev_up_lock);
				 return 0;
			}
    }
    spin_unlock(&pvfsdev_up_lock);
    return -EINVAL;
}


/* match_seq_entry()
 *
 * returns a pointer to the seq_entry with a matching sequence number,
 * or NULL if not found in the list.
 *
 * IMPORTANT! This function assumes that the caller holds a sequence
 * list semaphore when calling this function.
 */
static struct pvfsdev_seq_entry *match_seq_entry(struct list_head *my_list,
						 pvfs_seq_t my_seq)
{
    struct list_head *check = NULL;
    struct pvfsdev_seq_entry *entry = NULL;
    for (check = my_list->next; check != my_list; check = check->next) {
			entry = list_entry(check, struct pvfsdev_seq_entry, seq_list);

			if (entry->seq == my_seq) {
				 return entry;
			}
    }
    return NULL;
}

/* invalidate_seq_pending()
 *
 * this function will be used to set the invalid flag in a sequence
 * list entry. future operations on that sequence will be
 * aware that a previous step in the transfer has failed.
 *
 * IMPORTANT!  This function assumes that the caller is holding a
 * semaphore to protect the sequence list.
 *
 * Returns 0 on success, -errno on failure.
 */
static int invalidate_seq_pending(struct list_head *my_list,
				  pvfs_seq_t my_seq)
{
    struct list_head *check = NULL;
    struct pvfsdev_seq_entry *entry = NULL;
    for (check = my_list->next; check != my_list; check = check->next) {
			entry = list_entry(check, struct pvfsdev_seq_entry, seq_list);
			if (entry->seq == my_seq) {
				 entry->invalid = 1;
				 return 0;
			}
    }
    return -EINVAL;
}

/* is_fake_useraddr()
 *
 * This function tests for the case where someone has used the set_fs()
 * trick to pass a kernel address to a function that normally takes a
 * user address.
 *
 * Returns 0 if this appears to be a real user address, non-zero if not.
 */
static int is_fake_useraddr(void)
{
    mm_segment_t check_fs;
    mm_segment_t check_ds;

    check_fs = get_fs();
    check_ds = get_ds();

    if (check_fs.seg == check_ds.seg) {
			PDEBUG(D_SPECIAL, "is_fake_useraddr detected fake address\n");
			return 1;
    }
    return 0;
}

/*
 * Local variables:
 *  c-indent-level: 3
 *  c-basic-offset: 3
 *  tab-width: 3
 *
 * vim: ts=3
 * End:
 */ 
