/*
 * (C) 1995-2001 Clemson University and Argonne National Laboratory.
 *
 * See COPYING in top-level directory.
 */


#ifndef PVFS_CONFIG_H
#define PVFS_CONFIG_H


/* PVFS_CONFIG.H - SYSTEM CONFIGURATION PARAMETERS */

/* Release number:
 *
 * This is a base-10, 5 digit number, with one digit for the most
 * significant version number and two for the last two (e.g. 1.5.6 => 10506)
 */
#define PVFS_RELEASE_NR 10508

/* Capabilities:
 * These defines can be used to determine if the installed PVFS version
 * has certain capabilities:
 * HAVE_MGR_LOOKUP - mgr understands a lookup operation, which returns
 *                   most metadata but doesn't calculate file size
 * HAVE_PVFS_LISTIO - iods understand list i/o operations, and list i/o
 *                    functionality exists in the library.  this is used
 *                    to obtain higher performance in noncontiguous
 *                    access.
 * HAVE_MGR_CTIME - mgr understands a ctime operation, which can be used
 *                  modify the ctime on an existing file
 */
#define HAVE_MGR_LOOKUP
#define HAVE_PVFS_LISTIO
#define HAVE_MGR_CTIME

/* __USE_SENDFILE__ turns on the use of sendfile() in the I/O daemon.
 * We haven't seen any speedup from this yet, so it's also off by default.
 */
/* #undef  __USE_SENDFILE__ */

/* __RANDOM_NEXTSOCK__ turns on randomization in the socket selection
 * routines of nextsock(), defined in shared/sockset.c.  This hasn't
 * seemed to make a difference in any situations, so it's off by
 * default. This could probably never be turned on in the klibpvfs case.
 * mostly because the rand() stuff is still not in yet.
 */
/* #undef __RANDOM_NEXTSOCK__ */

/* __RANDOM_BASE__ turns on random selection of base nodes on the
 * manager.  This is VERY useful for groups that are using PVFS to store
 * large numbers of small files.  In the default case this sort of
 * situation will tend to result in I/O node 0 getting a lot of data
 * while the other nodes are mostly empty.
 *
 * On the other hand, it makes it more trouble to know where your data
 * is...and it makes it impossible to add more I/O nodes to an existing
 * configuration.
 */
/* #undef __RANDOM_BASE__ */

/* __ALWAYS_CONN__ turns on the O_CONN behavior all the time; sets up
 * connections to all I/O daemons when pvfs_open() is called.  This
 * makes the first read/write calls not include connection time, which
 * makes performance testing a little more straightforward.
 *
 * However, this feature has been found to expose a bug in the Intel
 * Etherexpress Pro fast ethernet card, causing it to hang and spit out
 * nasty messages.  So we've turned it off by default.
 */
/* #undef __ALWAYS_CONN__ */

/* __FS_H_IS_OK__  controls whether or not /linux/fs.h is included in the PVFS
 * manager and io daemon code.  One thing fs.h does is define NR_OPEN
 * to the kernel's default value, rather than pvfs's relatively small
 * default value of 256.  This controls how many files may be open
 * simultaneously by the file system.
 */
#define __FS_H_IS_OK__

/* PVFS_NR_OPEN controls how many open files a PVFS process can reliably
 * handle, including sockets and non-PVFS FDs.  This might need to be changed
 * if you get an error from pvfs_open().
 */
#define PVFS_NR_OPEN 1024

/* OUR FS MAGIC NUMBER...returned by statfs() */
#define PVFS_SUPER_MAGIC 0x0872

/* CLIENT_SOCKET_BUFFER_SIZE - clients will try to set send and receive
 * buffer sizes to this value
 */
#define CLIENT_SOCKET_BUFFER_SIZE (65535)

/* timeout value used when trying to connect to manager (see mgrcomm.c),
 * and when trying to brecv() the ack
 */
enum {
	MGRCOMM_CONNECT_TIMEOUT_SECS = 300,
	MGRCOMM_BRECV_TIMEOUT_SECS = 300,
	IODCOMM_CONNECT_TIMEOUT_SECS = 300,
	IODCOMM_BRECV_TIMEOUT_SECS = 300,
};

/* timeout used for brecv_timeout when trying to pull in requests and
 * acknowledgments.  This affects the iod and the mgr.
 */
enum {
	REQUEST_BRECV_TIMEOUT_SECS = 60
};

/* PVFS_LIST_IO_CUTOFF - maximum number of noncontiguous regions to use 
 * in a listio request
 */
enum {
	PVFS_LIST_IO_CUTOFF = 64
};

/* MANAGER STUFF */
#define MGR_REQ_PORT 3000
#define MAXIODS 512
#define MGR_BACKLOG 64

/* default stripe size */
#define DEFAULT_SSIZE 65536

/* not currently used... */
#define DEFAULT_BSIZE 0

/* inactivity timeout, ms */
#define SOCK_TIME_OUT 300000

/* TCP_CORK - should be easier to get; in /usr/include/linux/socket.h */
#ifndef TCP_CORK
#define TCP_CORK 3
#endif

/* SOL_TCP - not defined on IRIX */
#ifndef SOL_TCP
#define SOL_TCP 6
#endif

/* MAP_ANONYMOUS - not defined in IRIX */
#ifndef MAP_ANONYMOUS
#define MAP_ANONYMOUS 0
#endif

/* IOD STUFF */
#define IOD_REQ_PORT 7000
#define IOD_BASE_PORT IOD_REQ_PORT
#define IOD_BACKLOG 64

/* this determines how many directories the iod splits files into;
 * 101, 199, 499, 997 are all decent choices.
 */
#define IOD_DIR_HASH_SZ 101

#define IOD_USER "nobody"
#define IOD_GROUP "nobody"
#define IOD_ROOTDIR "/"
#define IOD_DATADIR "/pvfs-data"
#define IOD_LOGDIR "/tmp"

/* IOD_ACCESS_PAGE_SIZE - memory map region size (must be power of 2)
 *   - previously this value was 128K (for smaller machines)
 * IOD_WRITE_BUFFER_SIZE - size of buffer used for writing data
 *   - previously this value was 64K (for smaller machines)
 * IOD_SOCKET_BUFFER_SIZE - size of IOD socket buffers, recv and xmit
 *   - this value used to be 65535 by default, but admins can change it.
 */

#define IOD_ACCESS_PAGE_SIZE (512*1024)
#define IOD_WRITE_BUFFER_SIZE (512*1024)
#define IOD_SOCKET_BUFFER_SIZE (65535)

/* IOCTL DEFINES - COULDN'T FIND A BETTER PLACE TO PUT THEM... */
/* These are just arbitrary #s that linux doesn't seem to use. */
#define GETPART 0x5601
#define SETPART 0x5602
#define GETMETA 0x5603

/* more arbitrary #s; these are flags for pvfs_open() */
/* using lowest 4 bits of highest byte in the integer */
#define PVFSMASK 03700000000
#define O_META   00100000000
#define O_PART   00200000000
#define O_GSTART 00400000000
#define O_GADD   01000000000
#define O_CONN   02000000000


#endif
/*
 * Local variables:
 *  c-indent-level: 3
 *  c-basic-offset: 3
 *  tab-width: 3
 *
 * vim: ts=3
 * End:
 */ 
