/*
 * pvfs_proc.h
 * 
 * PVFS operation statistics. Based heavily on the CODA operation statistics file
 * See include/linux/coda_proc.h.
 *
 */

#ifndef _PVFS_PROC_H
#define _PVFS_PROC_H

void pvfs_sysctl_init(void);
void pvfs_sysctl_clean(void);
void pvfs_upcall_stats(int opcode, unsigned long jiffies);

#include "ll_pvfs.h"
#include <linux/sysctl.h>

/* these files are presented to show the result of the statistics:
 *
 *	/proc/fs/pvfs/vfs_stats
	 *		      upcall_stats
	 *		      io_traces
 *
 * these files are presented to reset the statistics to 0:
 *
 *	/proc/sys/pvfs/vfs_stats
 *		       upcall_stats
 *		       io_traces
 */

/* VFS operation statistics */

/* All the meta data operations */
struct pvfs_vfs_stats 
{
	/* file operations */
	int open;
	int lseek;
	int release;
	int mmap;
	int fsync;
	int read; 
	int write;

	/* dir operations */
	int readdir;
  
	/* inode operations */
	int create;
	int lookup;
	int link;
	int unlink;
	int mkdir;
	int rmdir;
	int rename;
	int truncate;
	int revalidate;
	int readpage;
};

/* This counts the number of upcalls and the rough amount of time spent
 * in the queues by the upcalls
 * Unfortunately floating point operations are not recommended inside
 * the kernel. Hence the unsigned long on the time quantities.
 */
struct pvfs_upcall_stats_entry 
{
  int count;
  unsigned long time_sum;
  unsigned long time_squared_sum;
};

struct pvfs_upcallstats {
	int total; /* total number of upcalls */
	int failures; /* total number of failed upcalls */
};

/* We essentially maintain a count of the read-write operations which fit under a certain bucket
 * size, What is exposed to the user-program,as a configurable parameter is the number of buckets,
 * individual bucket sizes. The maximum 
 */
enum {MAX_BUCKETS = 10, MAX_FILE_NAME_LENGTH = 128};

struct pvfs_rwop_stats {
	int nbuckets; /* nbuckets will be silently changed to MAX_BUCKETS if it is greater */
	int read_count[MAX_BUCKETS];
	int write_count[MAX_BUCKETS];
};

/* Update upcall stats from ll_pvfs.c */
void pvfs_upcall_stats(int opcode, long unsigned runtime);
extern struct pvfs_vfs_stats pvfs_vfs_stat;
extern struct pvfs_upcallstats pvfs_callstat;
void pvfs_rwop_stats(int rw, size_t size); 
extern int pvfs_upcall_timestamping, pvfs_collect_stats, pvfs_rwtrace_on;


/* reset statistics to 0 */
void reset_pvfs_vfs_stats( void );
void reset_pvfs_upcall_stats( void );
void reset_pvfs_rwop_stats( void );

void do_time_stats(struct pvfs_upcall_stats_entry * pentry, unsigned long jiffy );
unsigned long get_time_average( const struct pvfs_upcall_stats_entry * pentry );
unsigned long get_time_std_deviation( const struct pvfs_upcall_stats_entry * pentry );
int do_reset_pvfs_vfs_stats( ctl_table * table, int write, struct file * filp,
			     void * buffer, size_t * lenp );
int do_reset_pvfs_upcall_stats( ctl_table * table, int write, 
				struct file * filp, void * buffer, 
				size_t * lenp );
int do_reset_pvfs_rwop_stats(ctl_table *table, int write, 
				struct file *filp, void *buffer,
				size_t *lenp  );
/* these functions are called to form the content of /proc/fs/pvfs/... files */
int pvfs_vfs_stats_get_info( char * buffer, char ** start, off_t offset, int length);
int pvfs_upcall_stats_get_info( char * buffer, char ** start, off_t offset, int length);
int pvfs_rwop_stats_get_info(char *buffer, char **start, off_t offset, int length);

#endif /* _PVFS_PROC_H */
/*
 * Local variables:
 *  c-indent-level: 3
 *  c-basic-offset: 3
 *  tab-width: 3
 *
 * vim: ts=3
 * End:
 */ 
